import * as React from "react";
import {useEffect, useState} from "react";
import {observer} from "mobx-react-lite";
import {DashboardComponentModel} from "../../../../stores/DashboardComponentStore";
import {Table, TableBody, TableCell, TableHead, TableRow} from "@material-ui/core";
import {t} from "@lingui/macro";
import {ButtonField} from "@rx-suite/ui";
import {DashboardTargetStore} from "../../../../stores/DashboardTargetStore";
import {ScoreCardRowParameters} from "../ScoreCardRowParameters";
import {Parameter} from "../../../../stores/Types";
import {
    DragDropContext,
    Draggable,
    DraggingStyle,
    Droppable,
    DropResult,
    NotDraggingStyle,
    ResponderProvided
} from "react-beautiful-dnd";
import {observable} from "mobx";

export type TableSettingProps = {
    component: DashboardComponentModel,
}

export type DroppableComponentProps = {
    onDragEnd: (result: DropResult, provided: ResponderProvided) => void
}

export type DraggableComponentProps = {
    id: string,
    index: number
}

export type ScoreCardRowParameter = {
    row_name?: string,
    targets?: string[],
    goal?: number,
    goal_minimum?: number,
    row_type?: "throughput" | "machine_aborts" | "divider",
    data_operation?: string,
    target_number?: number,
    index?: number
}

const getItemStyle = (isDragging: boolean, draggableStyle: DraggingStyle | NotDraggingStyle | undefined) => ({
    // styles we need to apply on draggables
    ...draggableStyle,

    ...(isDragging && {
        background: "rgb(235,235,235)"
    })
})

const DraggableComponent: React.FunctionComponent<DraggableComponentProps> = observer(({id, index, ...props}) => {
    return (
        <Draggable draggableId={id} index={index}>
            {(provided, snapshot) => (
                <TableRow
                    ref={provided.innerRef}
                    {...provided.draggableProps}
                    {...provided.dragHandleProps}
                    style={getItemStyle(snapshot.isDragging, provided.draggableProps.style)}

                    {...props}
                >
                    {props.children}
                </TableRow>
            )}
        </Draggable>
    )
})

const DroppableComponent: React.FunctionComponent<DroppableComponentProps> = ({onDragEnd, ...props}) => {
    return (
        <DragDropContext onDragEnd={onDragEnd}>
            <Droppable droppableId={'1'} direction="vertical">
                {(provided) => {
                    return (
                        <TableBody ref={provided.innerRef} {...provided.droppableProps} {...props}>
                            {props.children}
                            {provided.placeholder}
                        </TableBody>
                    )
                }}
            </Droppable>
        </DragDropContext>
    )
}

export const ScoreCardSettings: React.FunctionComponent<TableSettingProps> = observer(({component}) => {
    const dashboardTarget = DashboardTargetStore.get(component.viewModel.dashboard_target_id);

    const maxRows = 28;
    // need to use a state, else everytime component updates (60 seconds) we lose our data of the viewModel.
    const [tableRows, setTableRows] = useState<ScoreCardRowParameter[]>([]);

    useEffect(() => {
        setTableRows(() => {
            if (!component.parameters) return [];
            if (!component.parameters.rows) return [];
            return component.parameters.rows.map((row: ScoreCardRowParameter) => observable(row));
        })
    }, [component])

    const addRowOnClick = React.useCallback(() => {
        const new_row = observable({row_name: "", index: tableRows.length}) as ScoreCardRowParameter;
        component.nestedAssignment('parameters.rows', [...tableRows, new_row]);
        setTableRows([...tableRows, new_row]);
    }, [tableRows, component])

    const onDragEnd = (result: DropResult) => {
        if (result.destination) {
            const rows = Array.from(tableRows);
            const [movedRow] = rows.splice(result.source.index, 1);
            rows.splice(result.destination.index, 0, movedRow);
            component.nestedAssignment('parameters.rows', rows);
            setTableRows(rows);
        }
    };

    const onDeleteRow = React.useCallback((rowIndex: number) => {
        const rows = Array.from(tableRows);
        rows.splice(rowIndex, 1);
        setTableRows(rows);
        component.nestedAssignment('parameters.rows', rows);
    }, [component, tableRows]);

    const headers = () => {
        const dynamicHeaders = Object.values(dashboardTarget?.parameters || {}).filter((param: Parameter) => param.single_row_parameter).map((param: Parameter) => {
            return <TableCell>{t`${param.name}`}</TableCell>
        });
        return dynamicHeaders.concat([
            <TableCell>{t`Minimum Target Value`}</TableCell>,
            <TableCell>{t`Target Value`}</TableCell>
        ]);
    }

    const getTableRows = React.useCallback(() => {
        return tableRows.map((tableRow, i) => {
            return <DraggableComponent key={`DraggableId ${i}`} id={`DraggableId ${i}`} index={i}>
                <ScoreCardRowParameters rowIndex={i}
                                        row={tableRow}
                                        component={component}
                                        dashboardTarget={dashboardTarget}
                                        onDeleteRow={onDeleteRow}
                />
            </DraggableComponent>
        })
    }, [component, dashboardTarget, tableRows, onDeleteRow])

    return <Table size={'small'}>
        <TableHead>
            <TableRow>
                {headers()}
            </TableRow>
        </TableHead>
        <DroppableComponent onDragEnd={onDragEnd}>
            {getTableRows()}
        </DroppableComponent>
        <ButtonField text={t`Add Table Row`} onClick={addRowOnClick}
                     disabled={tableRows.filter((row => row.row_type !== 'divider')).length >= maxRows}/>
    </Table>

})