import {createStyles, Divider, makeStyles, Theme, Typography} from "@material-ui/core";
import {t, Trans} from "@lingui/macro";
import * as React from "react";
import {observer} from "mobx-react-lite";
import {
    ButtonField,
    Conditional,
    IfFalse,
    IfTrue,
    InputField,
    ModelInput,
    ModelProvider,
    SelectField
} from "@rx-suite/ui";
import {Model} from "@rx-suite/data";
import {red} from "@material-ui/core/colors";
import {LabLayoutComponentModel} from "../../stores/LabLayoutComponentStore";
import {VisualAlertSettings} from "./VisualAlertSettings";
import {AlertGroupSelect} from "./AlertGroupSelect";
import {LayoutEditViewRenderer} from "../../stores/object/LayoutEditViewRenderer";
import {MachineStatusDotOptionsFields} from "./MachineStatusDotOptionsFields";

export type LayoutConfigComponentDetailsProps = {
    renderer: LayoutEditViewRenderer;
};

const useStyles = makeStyles((theme: Theme) =>
    createStyles({
        divider: {
            marginTop: theme.spacing(1),
            marginBottom: theme.spacing(1)
        },
        delete: {
            backgroundColor: red[500]
        }
    })
);

export const LayoutConfigComponentDetails: React.FunctionComponent<LayoutConfigComponentDetailsProps> = observer(({renderer}) => {
    const model = renderer.selectedComponent;
    const classes = useStyles();

    const componentSelectOptions = React.useCallback(() => {
        if (!renderer.layout) {
            return [];
        }
        return renderer.layout.current_components.reduce(function(result: Array<{label: string, value: LabLayoutComponentModel}>, c) {

            if (c.component_type !== 'status_legend') {

                result.push({label: c.name, value: c});

            } else if (c.viewModel.options && JSON.parse(c.viewModel.options).show) {

                // currently a status_legend component
                // hide component if no options, or show=false
                result.push({label: c.name, value: c});
            }

            return result;
        }, []);

    }, [renderer.layout]);

    const handleComponentSelect = React.useCallback((n: string | undefined, component: LabLayoutComponentModel) => {
        renderer.setSelectedComponent(component);
    }, [renderer]);

    const handleDeletedSelected = React.useCallback(() => {
        renderer.deleteSelectedComponent();
    }, [renderer]);

    const handleStatusDotWidthChange = React.useCallback((e) => {
        if (!model) return;
        const val = Number(e.target.value);
        model.viewModel.width = val;
        model.viewModel.height = val;
    }, [model, model?.viewModel.width, model?.viewModel.height]);

    const optionalFields = React.useMemo(() => {

        if (model?.viewModel.component_type === "machine_status_dot") {
            return <>
                <InputField
                    label={'Width'}
                    onChange={handleStatusDotWidthChange}
                    value={model ? model.viewModel.width : ''}
                />
                <Divider className={classes.divider}/>
                <MachineStatusDotOptionsFields model={model}/>
                <Divider className={classes.divider}/>
            </>;
        }

        if (model?.viewModel.component_type === "status_legend") {
            return <></>; // cannot change width/height
        }

        return <>
            <ModelInput name={'width'}/>
            <ModelInput name={'height'}/>
        </>;

    }, [classes.divider, handleStatusDotWidthChange, model]);

    const deleteComponent = React.useMemo(() => {

        if (model?.viewModel.component_type === "status_legend") {
            return <></>;
        }

        return <>
            <Divider className={classes.divider}/>
            <ButtonField
                text={t`Delete Component`}
                className={classes.delete}
                onClick={handleDeletedSelected}
            />
        </>;

    }, [classes.delete, classes.divider, handleDeletedSelected, model?.viewModel.component_type]);

    return (
        <React.Fragment>
            <ModelProvider value={model as Model}>
                <SelectField placeholder={t`Select component...`} onChange={handleComponentSelect}
                             value={model ? model : ""} options={componentSelectOptions()}/>
                <Divider className={classes.divider}/>
                <Conditional condition={!model}>
                    <IfTrue>
                        <Typography variant={'h6'}><Trans>No component selected.</Trans></Typography>
                    </IfTrue>
                    <IfFalse>
                        <ModelInput name={'x'}/>
                        <ModelInput name={'y'}/>
                        {optionalFields}
                        <ModelInput label={t`Layer`} name={'z_index'} type={'number'} inputProps={{min: 0, step: 1}}
                                    hint={{
                                        title: t`Layer`,
                                        text: t`Order in which components are placed on the layout. A component with greater layer is always in front of an element with a lower layer.`
                                    }}
                        />
                        <VisualAlertSettings model={model}/>
                        <AlertGroupSelect model={model}/>
                        {deleteComponent}
                    </IfFalse>
                </Conditional>
            </ModelProvider>
        </React.Fragment>
    )
});