import {action, computed, makeObservable, observable, reaction, when} from "mobx";
import {LabLayoutModel} from "../LabLayoutStore";
import {LabLayoutComponentModel, LayoutComponentPosition} from "../LabLayoutComponentStore";
import {anchorOption} from "../../components/lab_layout_edit/MachineStatusDotOptionsFields";

export abstract class CanvasRenderer {
    protected constructor() {
        makeObservable(this);
        reaction(() => this.selectedComponent?.viewModelAsJSON, () => {
            when(() => this.ctx, () => {
                this.draw();
            });
        });
    }

    @observable selectedComponent: LabLayoutComponentModel | undefined;

    @action.bound setSelectedComponent(c: LabLayoutComponentModel | undefined) {
        this.selectedComponent = c?.viewModel;
    }

    @observable state: LayoutEditState = 'done';
    protected lastMousePosition = CanvasRenderer.MOUSE_INITIAL_POSITION;

    @observable protected _layout: undefined | LabLayoutModel;
    @computed get layout(): LabLayoutModel {
        return (this._layout as unknown) as LabLayoutModel;
    }

    protected _canvas: HTMLCanvasElement | null = null;
    get canvas(): HTMLCanvasElement {
        return (this._canvas as unknown) as HTMLCanvasElement;
    }

    get ctx(): CanvasRenderingContext2D {
        return this.canvas.getContext('2d') as CanvasRenderingContext2D;
    }

    get ready(): boolean {
        return this._canvas != null && this._layout != undefined;
    }

    @action.bound
    setLayout(layout: LabLayoutModel): void {
        this.canvas.width = layout.resolution_x;
        this.canvas.height = layout.resolution_y;

        this._layout = layout;
        this.afterSetLayout();
        this.draw();
    }

    // provide hook for child when layout is set to update any data before a re-draw
    protected abstract afterSetLayout(): void;

    @action.bound
    setCanvasRef(c: any): void {
        if (c) {
            this._canvas = c;
        }
    }

    public draw(): void {
        if (!this.ready) return;

        this.clearCanvas();
        this.drawBorder();

        this.handleDraw();
    }

    // Let the child classes handle the drawing of the elements
    protected abstract handleDraw(): void;

    //// UTILITY FOR CANVAS SPACE CONVERSION

    protected convertToCanvasSpace(x: number, y: number): LayoutComponentPosition {
        const rect = this.canvas.getBoundingClientRect();
        x = x - rect.left;
        y = y - rect.top;
        const xScale = this.canvas.width / this.canvas.clientWidth;
        const yScale = this.canvas.height / this.canvas.clientHeight;
        return {
            x: x * xScale,
            y: y * yScale
        }
    }

    protected eventInputPosition(e: any): LayoutComponentPosition {
        if (e.touches) {
            return this.convertToCanvasSpace(e.touches[0].clientX, e.touches[0].clientY);
        }
        return this.convertToCanvasSpace(e.clientX, e.clientY);
    }

    //// DRAW FUNCTIONS

    protected drawStatusLegendTitleText(text: string, pos: LayoutComponentPosition): void {
        const scalingRatio = this.scaleRatio;
        const fontSize = 1.25 * scalingRatio.x;
        this.ctx.font = '600 ' + fontSize + 'rem Segoe UI';
        this.ctx.fillStyle = "black";
        this.ctx.textAlign = "start";
        this.ctx.fillText(text, pos.x, pos.y);
    }

    protected drawText(text: string, pos: LayoutComponentPosition): void {
        this.ctx.font = '600 ' + '0.75' + 'rem Segoe UI';
        this.ctx.fillStyle = "black";
        this.ctx.textAlign = "start";
        this.ctx.fillText(text, pos.x, pos.y);
    }

    protected drawMachineInfo(c: LabLayoutComponentModel, machineName: string, machineDuration: string, machineDotColor: string) {

        this.drawCircle({x: c.x, y: c.y}, c.width, {fillColor: machineDotColor});

        const defaultWidth = 120;

        const scalingRatio = (c.width / defaultWidth);

        const rectangleDimensions = {
            width: c.width,
            height: 30 * scalingRatio
        };

        const drawPositions = this.getMachineDotOptionPositions(c, c.options, rectangleDimensions);

        if (drawPositions.name) {
            const truncLength = 12;
            const truncMachineName = machineName.length > truncLength ? machineName.slice(0, truncLength) + '…' : machineName;
            this.drawRectangle(drawPositions.name.rectangle, rectangleDimensions.width, rectangleDimensions.height, { fillColor: 'white', strokeColor: 'black', lineWidth: 2 });
            this.drawMachineInfoText(truncMachineName, drawPositions.name.text, scalingRatio)
        }

        if (drawPositions.duration) {
            this.drawRectangle(drawPositions.duration.rectangle, rectangleDimensions.width, rectangleDimensions.height, { fillColor: 'white', strokeColor: 'black', lineWidth: 2 });
            this.drawMachineInfoText(machineDuration, drawPositions.duration.text, scalingRatio);
        }
    }

    private drawMachineInfoText(text: string, pos: LayoutComponentPosition, scalingRatio: number): void {
        const scaledFont = (18 * scalingRatio) + 'px';
        this.ctx.font = '600 ' + scaledFont + ' Segoe UI';
        this.ctx.fillStyle = "blue";
        this.ctx.textAlign = "center";
        this.ctx.fillText(text, pos.x, pos.y);
    }

    protected drawLine(start: LayoutComponentPosition, end: LayoutComponentPosition, opts: RenderOptions = {}) {
        this.ctx.save();
        this.setRenderOptions(opts);

        this.ctx.beginPath();
        this.ctx.moveTo(start.x, start.y)
        this.ctx.lineTo(end.x, end.y);
        this.ctx.stroke();
        this.ctx.closePath();

        this.ctx.restore();
    }

    protected drawRectangle(pos: LayoutComponentPosition, width: number, height: number, opts: RenderOptions = {}) {
        this.ctx.save();
        this.setRenderOptions(opts);

        this.ctx.beginPath();
        if (opts.fillColor) {
            if (opts.strokeColor) {
                this.ctx.fillRect(pos.x, pos.y, width, height);
                this.ctx.rect(pos.x, pos.y, width, height);
            } else {
                this.ctx.fillRect(pos.x, pos.y, width, height);
            }
        } else {
            this.ctx.rect(pos.x, pos.y, width, height);
        }
        this.ctx.stroke();
        this.ctx.closePath();

        this.ctx.restore();
    }

    protected drawTriangle(pos: LayoutComponentPosition, size: number, opts: RenderOptions = {}) {
        this.ctx.save();
        this.setRenderOptions(opts);

        this.ctx.translate(pos.x + size, pos.y);
        this.ctx.beginPath();
        this.ctx.moveTo(0, 0);
        this.ctx.lineTo(size, size * 2);
        this.ctx.lineTo(-size, size * 2);
        this.ctx.lineTo(0, 0);
        this.ctx.stroke();
        this.ctx.fill();
        this.ctx.closePath();

        this.ctx.restore();
    }

    protected drawExclamation(pos: LayoutComponentPosition, size: number, opts: RenderOptions = {}) {
        this.ctx.save();
        this.setRenderOptions(opts);

        this.ctx.translate(pos.x + size, pos.y + (size / 4));
        this.ctx.beginPath();
        this.ctx.arc(0, 0, size / 4, 0, Math.PI, true);
        this.ctx.arc(0, size + (size / 4), size / 9, Math.PI, 0, true);
        this.ctx.lineTo(size / 4, 0);
        this.ctx.stroke();
        this.ctx.fill();
        this.ctx.closePath();
        this.ctx.beginPath();
        this.ctx.arc(0, size + size / 1.6, size / 8, 0, Math.PI * 2);
        this.ctx.stroke();
        this.ctx.fill();

        this.ctx.restore();
    }

    protected drawCircle(pos: LayoutComponentPosition, size: number, opts: RenderOptions = {}): void {
        this.ctx.save();
        this.setRenderOptions(opts);

        this.ctx.beginPath();
        this.ctx.arc(pos.x + (size / 2), pos.y + (size / 2), size / 2, 0, Math.PI * 2, false);
        this.ctx.fill();
        this.ctx.closePath();

        this.ctx.restore();
    }

    protected drawImage(c: LabLayoutComponentModel): void {
        // if the image is already loaded, render it now
        if (c.image) {
            this.ctx.drawImage(c.image, c.x, c.y, c.width, c.height)
            // Image was already saved with the back end, load it from there
        } else if (c.image_url) {
            c.image = new Image();
            c.image.src = c.image_url;
            c.image.onload = () => this.draw();
            // Image exists only in the front-end, load it from a blob
        } else {
            const reader = new FileReader();
            reader.onload = (b: ProgressEvent<FileReader>) => {
                if (!b.target) return;
                c.image = new Image();
                c.image.src = b.target.result as string;
                c.image.onload = () => {
                    if (!c.image) return
                    c.width = c.image.width as number;
                    c.height = c.image.height as number;
                    c.aspect_ratio = c.width / c.height;
                }
            };
            if (c.file) {
                reader.readAsDataURL(c.file);
            }
        }
    }

    protected drawAlertGroup(c: LabLayoutComponentModel, opts: RenderOptions = {}): void {
        const fillColor = opts.fillColor ? opts.fillColor : CanvasRenderer.COLOR_ALERT_OFF;
        const triangleSize = c.height / 4;
        const exclamationSize = triangleSize / 2;
        this.drawRectangle(c.position, c.width, c.height, {strokeColor: fillColor, lineWidth: 7});
        this.drawTriangle({
            x: (c.position.x + c.width / 2) - triangleSize,
            y: c.position.y + triangleSize
        }, triangleSize, {strokeColor: fillColor, fillColor: fillColor})
        this.drawExclamation({
            x: (c.position.x + c.width / 2) - exclamationSize,
            y: c.position.y + triangleSize * 1.75
        }, exclamationSize, {strokeColor: 'white', fillColor: 'white'});
    }

    protected getMachineDotOptionPositions(c: LabLayoutComponentModel, options: string | null, rDimensions: { width: number, height: number }) {

        if (!options) {
            return {
                name: null,
                duration: null
            };
        }

        const optionsJson: { showName: anchorOption | null, showDuration: anchorOption | null } = JSON.parse(options);

        const singlePositions = {
            top: {
                rectangle: {x: c.position.x, y: c.position.y - (5 * rDimensions.height / 4)},
                text: {x: c.position.x + (c.width / 2), y: c.position.y - (rDimensions.height / 2)}
            },
            right: {
                rectangle: {x: c.position.x + c.width + 15, y: c.position.y + (c.height / 2) - (rDimensions.height / 2)},
                text: {x: c.position.x + (3 * c.width / 2) + 15, y: c.position.y + (c.height / 2) + (rDimensions.height / 4)}
            },
            left: {
                rectangle: {x: c.position.x - (c.width) - 15, y: c.position.y + (c.height / 2) - (rDimensions.height / 2)},
                text: {x: c.position.x - (c.width / 2) - 15, y: c.position.y + (c.height / 2) + (rDimensions.height / 4)}
            },
            bottom: {
                rectangle: {x: c.position.x, y: c.position.y + c.height + (rDimensions.height / 4)},
                text: {x: c.position.x + (c.width / 2), y: c.position.y + c.height + rDimensions.height}
            }
        };

        if (optionsJson.showName === optionsJson.showDuration) {

            if (optionsJson.showName == null) { // both are null, return safely
                return {
                    name: null,
                    duration: null
                };
            }

            const spacingForTopBottom = (5 * rDimensions.height / 4);
            const spacingForLeftRight = (rDimensions.height / 2) + 3.75;

            // Get stacking positions

            const positions = {
                top: [
                    {
                        rectangle: {x: singlePositions.top.rectangle.x, y: singlePositions.top.rectangle.y - spacingForTopBottom},
                        text: {x: singlePositions.top.text.x, y: singlePositions.top.text.y - spacingForTopBottom}
                    },
                    singlePositions.top
                ],
                right: [
                    {
                        rectangle: {x: singlePositions.right.rectangle.x, y: singlePositions.right.rectangle.y - spacingForLeftRight},
                        text: {x: singlePositions.right.text.x, y: singlePositions.right.text.y - spacingForLeftRight}
                    },
                    {
                        rectangle: {x: singlePositions.right.rectangle.x, y: singlePositions.right.rectangle.y + spacingForLeftRight},
                        text: {x: singlePositions.right.text.x, y: singlePositions.right.text.y + spacingForLeftRight}
                    }
                ],
                left: [
                    {
                        rectangle: {x: singlePositions.left.rectangle.x, y: singlePositions.left.rectangle.y - spacingForLeftRight},
                        text: {x: singlePositions.left.text.x, y: singlePositions.left.text.y - spacingForLeftRight}
                    },
                    {
                        rectangle: {x: singlePositions.left.rectangle.x, y: singlePositions.left.rectangle.y + spacingForLeftRight},
                        text: {x: singlePositions.left.text.x, y: singlePositions.left.text.y + spacingForLeftRight}
                    }
                ],
                bottom: [
                    singlePositions.bottom,
                    {
                        rectangle: {x: singlePositions.bottom.rectangle.x, y: singlePositions.bottom.rectangle.y + spacingForTopBottom},
                        text: {x: singlePositions.bottom.text.x, y: singlePositions.bottom.text.y + spacingForTopBottom}
                    }
                ]
            }[optionsJson.showName];

            return {
                name: positions[0],
                duration: positions[1]
            };
        }

        return {
            name: optionsJson.showName ? singlePositions[optionsJson.showName] : null,
            duration: optionsJson.showDuration ? singlePositions[optionsJson.showDuration] : null
        };
    }

    protected get scaleRatio(): { x: number, y: number } {

        const defaultResolution = { x: 1920, y: 1080 };

        const currentLayout = this.layout;
        if (!currentLayout) {
            return { x: 1, y: 1 };
        }

        return {
            x: currentLayout.resolution_x / defaultResolution.x,
            y: currentLayout.resolution_y / defaultResolution.y
        };
    }


    private clearCanvas(): void {
        this.ctx.clearRect(0, 0, this.canvas.width, this.canvas.height);
    }

    private setRenderOptions(opts: RenderOptions) {
        if (opts.fillColor) this.ctx.fillStyle = opts.fillColor;
        if (opts.lineWidth) this.ctx.lineWidth = opts.lineWidth;
        if (opts.strokeColor) this.ctx.strokeStyle = opts.strokeColor;
    }

    private drawBorder(): void {
        this.drawRectangle({x: 0, y: 0}, this.canvas.width, this.canvas.height);
    }

    protected static COLOR_ALERT_OFF = "#FFEBEE";
    protected static COLOR_ALERT_ON = "#B71C1C";

    protected static MOUSE_INITIAL_POSITION = {x: 0, y: 0};
}

type RenderOptions = {
    strokeColor?: string;
    lineWidth?: number;
    fillColor?: string;
}


type LayoutEditState = 'done' | 'mouse_down' | 'dragging' | 'saving' | 'resizing';